@@ -17,11 +17,22 @@ package IO::Socket::SSL;
 use IO::Socket;
 use Net::SSLeay 1.21;
 use Exporter ();
-use Scalar::Util 'dualvar';
-use Errno 'EAGAIN';
+use Errno qw( EAGAIN ETIMEDOUT );
 use Carp;
 use strict;
 
+# from openssl/ssl.h, should be better in Net::SSLeay
+use constant SSL_SENT_SHUTDOWN => 1;
+use constant SSL_RECEIVED_SHUTDOWN => 2;
+
+# non-XS Versions of Scalar::Util will fail
+BEGIN{
+    eval { use Scalar::Util 'dualvar'; dualvar(0,'') };
+    die "You need the XS Version of Scalar::Util for dualvar() support" 
+	if $@;
+}
+
+
 use vars qw(@ISA $VERSION $DEBUG $SSL_ERROR $GLOBAL_CONTEXT_ARGS @EXPORT );
 
 {
@@ -41,7 +52,7 @@ use vars qw(@ISA $VERSION $DEBUG $SSL_ERROR $GLOBAL_CONTEXT_ARGS @EXPORT );
 BEGIN {
     # Declare @ISA, $VERSION, $GLOBAL_CONTEXT_ARGS
     @ISA = qw(IO::Socket::INET);
-    $VERSION = '1.05';
+    $VERSION = '1.12';
     $GLOBAL_CONTEXT_ARGS = {};
 
     #Make $DEBUG another name for $Net::SSLeay::trace
@@ -57,6 +68,14 @@ BEGIN {
 
 }
 
+sub DEBUG {
+    $DEBUG or return;
+    my (undef,undef,$line) = caller;
+    my $msg = shift;
+    $msg = sprintf $msg,@_ if @_;
+    print STDERR "DEBUG: $line: $msg\n";
+}
+
 sub CLONE_SKIP { 1 }
 
 # Export some stuff
@@ -139,7 +158,7 @@ sub configure_SSL {
     %$arg_hash = ( %default_args, %$GLOBAL_CONTEXT_ARGS, %$arg_hash );
 
     #Avoid passing undef arguments to Net::SSLeay
-    !defined($arg_hash->{$_}) and ($arg_hash->{$_} = '') foreach (keys %$arg_hash);
+    defined($arg_hash->{$_}) or delete($arg_hash->{$_}) foreach (keys %$arg_hash);
 
     #Handle CA paths properly if no CA file is specified
     if ($arg_hash->{'SSL_ca_path'} ne '' and !(-f $arg_hash->{'SSL_ca_file'})) {
@@ -189,6 +208,7 @@ sub connect {
 
 sub connect_SSL {
     my $self = shift;
+    my $args = @_>1 ? {@_}: $_[0]||{};
 
     my ($ssl,$ctx);
     if ( ! ${*$self}{'_SSL_opening'} ) {
@@ -220,29 +240,82 @@ sub connect_SSL {
     $ssl ||= ${*$self}{'_SSL_object'};
 
     $SSL_ERROR = undef;
-    #DEBUG( 'calling ssleay::connect' );
-    my $rv = Net::SSLeay::connect($ssl);
-    #DEBUG( "rv=$rv" );
-    if ( $rv < 0 ) {
-	unless ( $self->_set_rw_error( $ssl,$rv )) {
-	    $self->error("SSL connect attempt failed with unknown error");
+    my $timeout = exists $args->{Timeout} 
+    	? $args->{Timeout} 
+	: ${*$self}{io_socket_timeout}; # from IO::Socket
+    if ( defined($timeout) && $timeout>0 && $self->blocking(0) ) {
+	#DEBUG( "set socket to non-blocking to enforce timeout=$timeout" );
+	# timeout was given and socket was blocking
+    	# enforce timeout with now non-blocking socket
+    } else {
+	# timeout does not apply because invalid or socket non-blocking
+    	$timeout = undef; 
+    }
+
+    my $start = defined($timeout) && time();
+    for my $dummy (1) {
+	#DEBUG( 'calling ssleay::connect' );
+	my $rv = Net::SSLeay::connect($ssl);
+	#DEBUG( "connect -> rv=$rv" );
+	if ( $rv < 0 ) {
+	    unless ( $self->_set_rw_error( $ssl,$rv )) {
+		$self->error("SSL connect attempt failed with unknown error");
+		delete ${*$self}{'_SSL_opening'};
+		${*$self}{'_SSL_opened'} = 1;
+		#DEBUG( "fatal SSL error: $SSL_ERROR" );
+		return $self->fatal_ssl_error();
+	    }
+
+	    #DEBUG( 'ssl handshake in progress' );
+	    # connect failed because handshake needs to be completed
+	    # if socket was non-blocking or no timeout was given return with this error
+	    return if ! defined($timeout);
+
+	    # wait until socket is readable or writable
+	    my $rv;
+	    if ( $timeout>0 ) {
+		my $vec = '';
+		vec($vec,$self->fileno,1) = 1;
+		#DEBUG( "waiting for fd to become ready: $SSL_ERROR" );
+	    	$rv = 
+		    $SSL_ERROR == SSL_WANT_READ ? select( $vec,undef,undef,$timeout) :
+		    $SSL_ERROR == SSL_WANT_WRITE ? select( undef,$vec,undef,$timeout) :
+		    undef;
+	    } else {
+		#DEBUG( "handshake failed because no more time" );
+	    	$! = ETIMEDOUT
+	    }
+	    if ( ! $rv ) {
+		#DEBUG( "handshake failed because socket did not became ready" );
+		# failed because of timeout, return
+	    	$! ||= ETIMEDOUT;
+		delete ${*$self}{'_SSL_opening'};
+		${*$self}{'_SSL_opened'} = 1;
+		$self->blocking(1); # was blocking before
+	    	return 
+	    }
+
+	    # socket is ready, try non-blocking connect again after recomputing timeout
+	    #DEBUG( "socket ready, retrying connect" );
+	    my $now = time();
+	    $timeout -= $now - $start;
+	    $start = $now;
+	    redo;
+
+	} elsif ( $rv == 0 ) {
 	    delete ${*$self}{'_SSL_opening'};
+	    #DEBUG( "connection failed - connect returned 0" );
+	    $self->error("SSL connect attempt failed because of handshake problems" );
 	    ${*$self}{'_SSL_opened'} = 1;
 	    return $self->fatal_ssl_error();
 	}
-	#DEBUG( 'ssl handshake in progress' );
-	return;
-    } elsif ( $rv == 0 ) {
-	delete ${*$self}{'_SSL_opening'};
-	$self->error("SSL connect attempt failed because of handshake problems" );
-	${*$self}{'_SSL_opened'} = 1;
-	return $self->fatal_ssl_error();
     }
 
     #DEBUG( 'ssl handshake done' );
     # ssl connect successful
     delete ${*$self}{'_SSL_opening'};
     ${*$self}{'_SSL_opened'}=1;
+    $self->blocking(1) if defined($timeout); # was blocking before
 
     $ctx ||= ${*$self}{'_SSL_ctx'};
     if ( $ctx->has_session_cache ) {
@@ -292,8 +365,9 @@ sub accept {
 }
 
 sub accept_SSL {
-    my ($self,$socket) = @_;
-    $socket ||= $self;
+    my $self = shift;
+    my $socket = ( @_ && UNIVERSAL::isa( $_[0], 'IO::Handle' )) ? shift : $self;
+    my $args = @_>1 ? {@_}: $_[0]||{};
 
     my $ssl;
     if ( ! ${*$self}{'_SSL_opening'} ) {
@@ -322,27 +396,74 @@ sub accept_SSL {
 
     $SSL_ERROR = undef;
     #DEBUG( 'calling ssleay::accept' );
-    my $rv = Net::SSLeay::accept($ssl);
-    #DEBUG( 'called ssleay::accept rv='.$rv );
-    if ( $rv < 0 ) {
-	unless ( $socket->_set_rw_error( $ssl,$rv )) {
-	    $socket->error("SSL accept attempt failed with unknown error");
+
+    my $timeout = exists $args->{Timeout} 
+    	? $args->{Timeout} 
+	: ${*$self}{io_socket_timeout}; # from IO::Socket
+    if ( defined($timeout) && $timeout>0 && $socket->blocking(0) ) {
+	# timeout was given and socket was blocking
+    	# enforce timeout with now non-blocking socket
+    } else {
+	# timeout does not apply because invalid or socket non-blocking
+    	$timeout = undef; 
+    }
+
+    my $start = defined($timeout) && time();
+    for my $dummy (1) {
+	my $rv = Net::SSLeay::accept($ssl);
+	#DEBUG( 'called ssleay::accept rv='.$rv );
+	if ( $rv < 0 ) {
+	    unless ( $socket->_set_rw_error( $ssl,$rv )) {
+		$socket->error("SSL accept attempt failed with unknown error");
+		delete ${*$self}{'_SSL_opening'};
+		${*$socket}{'_SSL_opened'} = 1;
+		return $socket->fatal_ssl_error();
+	    }
+
+	    # accept failed because handshake needs to be completed
+	    # if socket was non-blocking or no timeout was given return with this error
+	    return if ! defined($timeout);
+
+	    # wait until socket is readable or writable
+	    my $rv;
+	    if ( $timeout>0 ) {
+		my $vec = '';
+		vec($vec,$socket->fileno,1) = 1;
+	    	$rv = 
+		    $SSL_ERROR == SSL_WANT_READ ? select( $vec,undef,undef,$timeout) :
+		    $SSL_ERROR == SSL_WANT_WRITE ? select( undef,$vec,undef,$timeout) :
+		    undef;
+	    } else {
+	    	$! = ETIMEDOUT
+	    }
+	    if ( ! $rv ) {
+		# failed because of timeout, return
+	    	$! ||= ETIMEDOUT;
+		delete ${*$self}{'_SSL_opening'};
+		${*$socket}{'_SSL_opened'} = 1;
+		$socket->blocking(1); # was blocking before
+	    	return 
+	    }
+
+	    # socket is ready, try non-blocking accept again after recomputing timeout
+	    my $now = time();
+	    $timeout -= $now - $start;
+	    $start = $now;
+	    redo;
+
+	} elsif ( $rv == 0 ) {
+	    $socket->error("SSL connect accept failed because of handshake problems" );
 	    delete ${*$self}{'_SSL_opening'};
-    	    ${*$socket}{'_SSL_opened'} = 1;
+	    ${*$socket}{'_SSL_opened'} = 1;
 	    return $socket->fatal_ssl_error();
 	}
-	return;
-    } elsif ( $rv == 0 ) {
-	$socket->error("SSL connect accept failed because of handshake problems" );
-	delete ${*$self}{'_SSL_opening'};
-	${*$socket}{'_SSL_opened'} = 1;
-	return $socket->fatal_ssl_error();
     }
 
     #DEBUG( 'handshake done, socket ready' );
     # socket opened
     delete ${*$self}{'_SSL_opening'};
     ${*$socket}{'_SSL_opened'} = 1;
+    $socket->blocking(1) if defined($timeout); # was blocking before
 
     tie *{$socket}, "IO::Socket::SSL::SSL_HANDLE", $socket;
 
@@ -537,33 +658,94 @@ sub readline {
 sub close {
     my $self = shift || return _invalid_object();
     my $close_args = (ref($_[0]) eq 'HASH') ? $_[0] : {@_};
+
+    return if ! $self->stop_SSL(
+	SSL_fast_shutdown => 1,
+	%$close_args,
+	_SSL_ioclass_downgrade => 0,
+    );
+
+    if ( ! $close_args->{_SSL_in_DESTROY} ) {
+	untie( *$self );
+    	return $self->SUPER::close;
+    }
+    return 1;
+}
+
+sub stop_SSL {
+    my $self = shift || return _invalid_object();
+    my $stop_args = (ref($_[0]) eq 'HASH') ? $_[0] : {@_};
     return $self->error("SSL object already closed") unless (${*$self}{'_SSL_opened'});
 
     if (my $ssl = ${*$self}{'_SSL_object'}) {
-	local $SIG{PIPE} = sub{};
-	$close_args->{'SSL_no_shutdown'} or Net::SSLeay::shutdown($ssl);
+	my $shutdown_done;
+	if ( $stop_args->{SSL_no_shutdown} ) {
+	    $shutdown_done = 1;
+	} else {
+	    my $fast = $stop_args->{SSL_fast_shutdown};
+	    my $status = Net::SSLeay::get_shutdown($ssl);
+	    if ( $status == SSL_RECEIVED_SHUTDOWN 
+	    	|| ( $status != 0 && $fast )) {
+	    	# shutdown done
+	    	$shutdown_done = 1;
+	    } else {
+		# need to initiate/continue shutdown
+	    	local $SIG{PIPE} = sub{};
+		for my $try (1,2 ) {
+		    my $rv = Net::SSLeay::shutdown($ssl);
+		    if ( $rv < 0 ) {
+			# non-blocking socket?
+			$self->_set_rw_error( $ssl,$rv );
+			# need to try again
+			return;
+		    } elsif ( $rv
+			|| ( $rv == 0 && $fast )) {
+			# shutdown finished
+	    		$shutdown_done = 1;
+			last;
+		    } else {
+			# shutdown partly finished (e.g. one direction)
+			# call again
+		    }
+		}
+	    }
+	}
+
+	return if ! $shutdown_done;
 	Net::SSLeay::free($ssl);
-	delete ${*$self}{'_SSL_object'};
+	delete ${*$self}{_SSL_object};
     }
 
-    if ($close_args->{'SSL_ctx_free'}) {
-	my $ctx = ${*$self}{'_SSL_ctx'};
-	delete ${*$self}{'_SSL_ctx'};
-	$ctx->DESTROY();
+    if ($stop_args->{'SSL_ctx_free'}) {
+	my $ctx = delete ${*$self}{'_SSL_ctx'};
+	$ctx && $ctx->DESTROY();
     }
 
-    if (${*$self}{'_SSL_certificate'}) {
-	Net::SSLeay::X509_free(${*$self}{'_SSL_certificate'});
+    if (my $cert = delete ${*$self}{'_SSL_certificate'}) {
+	Net::SSLeay::X509_free($cert);
     }
 
     ${*$self}{'_SSL_opened'} = 0;
-    my $arg_hash = ${*$self}{'_SSL_arguments'};
-    untie(*$self) unless ($arg_hash->{'SSL_server'}
-			  or $close_args->{_SSL_in_DESTROY});
 
-    $self->SUPER::close unless ($close_args->{_SSL_in_DESTROY});
+    if ( ! $stop_args->{_SSL_in_DESTROY} ) {
+
+	my $downgrade = $stop_args->{_SSL_ioclass_downgrade};
+	if ( $downgrade || ! defined $downgrade ) {
+	    # rebless to original class from start_SSL
+	    if ( my $orig_class = delete ${*$self}{'_SSL_ioclass_upgraded'} ) {
+		bless $self,$orig_class;
+		untie(*$self);
+		# FIXME: if original class was tied too we need to restore the tie
+	    }
+	    # remove all _SSL related from *$self
+	    my @sslkeys = grep { m{^_?SSL_} } keys %{*$self};
+	    delete @{*$self}{@sslkeys} if @sslkeys;
+	}
+    }
+    return 1;
 }
 
+
 sub kill_socket {
     my $self = shift;
     shutdown($self, 2);
@@ -603,6 +785,7 @@ sub start_SSL {
     my ($class,$socket) = (shift,shift);
     return $class->error("Not a socket") unless(ref($socket));
     my $arg_hash = (ref($_[0]) eq 'HASH') ? $_[0] : {@_};
+    my %to = exists $arg_hash->{Timeout} ? ( Timeout => delete $arg_hash->{Timeout} ) :();
     my $original_class = ref($socket);
     my $original_fileno = (UNIVERSAL::can($socket, "fileno"))
 	? $socket->fileno : CORE::fileno($socket);
@@ -612,6 +795,7 @@ sub start_SSL {
     $socket->configure_SSL($arg_hash) or bless($socket, $original_class) && return;
 
     ${*$socket}{'_SSL_fileno'} = $original_fileno;
+    ${*$socket}{'_SSL_ioclass_upgraded'} = $original_class;
 
     my $start_handshake = $arg_hash->{SSL_startHandshake};
     if ( ! defined($start_handshake) || $start_handshake ) {
@@ -619,8 +803,8 @@ sub start_SSL {
 	#DEBUG( "start handshake" );
 	my $blocking = $socket->blocking(1);
 	my $result = ${*$socket}{'_SSL_arguments'}{SSL_server}
-	    ? $socket->accept_SSL
-	    : $socket->connect_SSL;
+	    ? $socket->accept_SSL(%to)
+	    : $socket->connect_SSL(%to);
 	$socket->blocking(0) if !$blocking;
     	return $result ? $socket : (bless($socket, $original_class) && ());
     } else {
@@ -684,6 +868,7 @@ sub errstr {
 sub fatal_ssl_error {
     my $self = shift;
     my $error_trap = ${*$self}{'_SSL_arguments'}->{'SSL_error_trap'};
+    $@ = $self->errstr;
     if (defined $error_trap and ref($error_trap) eq 'CODE') {
 	$error_trap->($self, $self->errstr()."\n".$self->get_ssleay_error());
     } else { $self->kill_socket; }
@@ -1076,6 +1261,8 @@ see the note about return values).
 If you are using non-blocking sockets read on, as version 0.98 added better
 support for non-blocking.
 
+If you are trying to use it with threads see the BUGS section.
+
 =head1 METHODS
 
 IO::Socket::SSL inherits its methods from IO::Socket::INET, overriding them
@@ -1268,6 +1455,13 @@ If set to a true value, this option will make close() not use the SSL_shutdown()
 on the socket in question so that the close operation can complete without problems
 if you have used shutdown() or are working on a copy of a socket.
 
+=item SSL_fast_shutdown
+
+If set to true only a unidirectional shutdown will be done, e.g. only the 
+close_notify (see SSL_shutdown(3)) will be called. Otherwise a bidrectional
+shutdown will be done. If used within close() it defaults to true, if used
+within stop_SSL() it defaults to false.
+
 =item SSL_ctx_free
 
 If you want to make sure that the SSL context of the socket is destroyed when
@@ -1334,6 +1528,21 @@ IO::Socket::SSL and call accept_SSL or connect_SSL and the upgraded object. To
 just upgrade the socket set B<SSL_startHandshake> explicitly to 0. If you call start_SSL
 w/o this parameter it will revert to blocking behavior for accept_SSL and connect_SSL.
 
+If given the parameter "Timeout" it will stop if after the timeout no SSL connection
+was established. This parameter is only used for blocking sockets, if it is not given the
+default Timeout from the underlying IO::Socket will be used.
+
+=item B<stop_SSL(...)>
+
+This is the opposite of start_SSL(), e.g. it will shutdown the SSL connection
+and return to the class before start_SSL(). It gets the same arguments as close(),
+in fact close() calls stop_SSL() (but without downgrading the class).
+
+Will return true if it suceeded and undef if failed. This might be the case for
+non-blocking sockets. In this case $! is set to EAGAIN and the ssl error to
+SSL_WANT_READ or SSL_WANT_WRITE. In this case the call should be retried again with 
+the same arguments once the socket is ready is until it succeeds.
+
 =item B<< IO::Socket::SSL->new_from_fd($fd, ...) >>
 
 This will convert a socket identified via a file descriptor into an SSL socket.
@@ -1459,6 +1668,8 @@ IO::Socket::SSL is not threadsafe.
 This is because IO::Socket::SSL is based on Net::SSLeay which 
 uses a global object to access some of the API of openssl
 and is therefore not threadsafe.
+It might probably work if you don't use SSL_verify_cb and
+SSL_password_cb.
 
 IO::Socket::SSL does not work together with Storable::fd_retrieve/fd_store.
 See BUGS file for more information and how to work around the problem.
